# Part of atomate2-turbomole package.

"""Module containing the custodian validators for the atomate2-turbomole package."""

import os
from dataclasses import dataclass

from custodian.custodian import Validator
from turbomoleio.output.data import ScfIterationData


@dataclass
class JobexGeoOptConvergedValidator(Validator):
    """Validator for jobex run.

    Validates that a GEO_OPT_CONVERGED file has been written.
    """

    def check(self, directory="./"):
        """Check for jobex convergence error.

        Returns
        -------
            bool: False if jobex is converged, i.e. a GEO_OPT_CONVERGED file
                is present in the directory, True otherwise
        """
        return not os.path.exists("GEO_OPT_CONVERGED")


@dataclass
class ScfConvergedValidator(Validator):
    """Validator for scf run (ridft or dscf).

    Validates the scf run is converged.
    """

    output_file: str

    def check(self, directory="./"):
        """Check for scf convergence error.

        Returns
        -------
            bool: False if scf run is converged.
        """
        if not os.path.exists(self.output_file):
            return True
        scf_iterations = ScfIterationData.from_file(self.output_file)
        if scf_iterations:
            return not scf_iterations.converged
        # For riper calculations, turbomoleio does not get scf iterations (yet)
        # Just make sure it is converged by reading the output file directly
        with open(self.output_file) as f:
            string = f.read()
        return "SCF converged within " not in string


@dataclass
class NormalEndValidator(Validator):
    """Check for the presence of the "all done" string in the output file."""

    out_file: str

    def check(self, directory="./"):
        """
        Check normal end.

        Return True if an error is found, i.e. output file is not present or
        the "all done" string is not present in the output file.
        """
        if not os.path.isfile(self.out_file):
            return True

        # read bytes since sometimes error file could be corrupted,
        # this will avoid failures due to non utf-8 bytes.
        with open(self.out_file, "rb") as f:
            out = f.read()

        failed = out.rfind(b"all done") < 0

        return failed



# Part of atomate2-turbomole package.

# Automatically get the version of the package
# For python versions < 3.8, importlib.metadata module did not exist,
# had to use importlib_metadata instead.
try:  # pragma: no cover
    from importlib.metadata import version
except ModuleNotFoundError:  # pragma: no cover
    from importlib_metadata import version  # type: ignore

__version__ = version("atomate2-turbomole")



# # Part of turbomole package.
#
# """Tests of the custodian jobs in the turbomole package."""
#
# import subprocess
#
# import pytest
# from monty.tempfile import ScratchDir
#
# import turbomole.custodian_jobs
# from turbomole.custodian_jobs import Dscf, Jobex, Ricc2, Ridft, Riper, Statpt
#
#
# def test_jbx_job(mocker):
#     mock = mocker.patch.object(turbomole.custodian_jobs, "subprocess", autospec=True)
#     job = Jobex()
#     assert job.options == ["-time"]
#     job = Jobex(jobex_time=True, options=["-time"])
#     assert job.options == ["-time"]
#     job = Jobex(jobex_time=True, options="-time")
#     assert job.options == ["-time"]
#     job = Jobex(options="-rijk")
#     assert "-time" in job.options
#     assert "-rijk" in job.options
#     with ScratchDir("."):
#         p = job.run()
#         assert isinstance(p, subprocess.Popen)
#     mock.Popen.assert_called_once()
#
#     job = Jobex(jobex_time=False, options=["-level", "cc2"])
#     assert "-time" not in job.options
#     assert "-level" in job.options
#     assert "cc2" in job.options
#
#     with pytest.raises(TypeError):
#         Jobex(options=5)
#
#     for jb_cls in [
#         Ridft,
#         Riper,
#         Dscf,
#         Ricc2,
#         Statpt,
#     ]:
#         mock.reset_mock()
#         jb = jb_cls()
#         with ScratchDir("."):
#             p = jb.run()
#             assert isinstance(p, subprocess.Popen)
#         mock.Popen.assert_called_once()
#
#         jb = jb_cls(options="-proper")
#         assert len(jb.options) == 1
#         assert "-proper" in jb.options
#
#         jb = jb_cls(options=["-opt", "optabc"])
#         assert len(jb.options) == 2
#         assert "-opt" in jb.options
#         assert "optabc" in jb.options
#
#         with pytest.raises(TypeError):
#             jb_cls(options=5)



# Part of atomate2-turbomole package.

"""Tests of the utility functions in the atomate2-turbomole package."""

import os

import pytest
from atomate2.turbomole.utils import JobexTimings, get_define_parameters


def test_jbx_timings(test_files):
    # time.stat file from "jobex -riper -time" (relaxation with riper,
    # i.e. periodic system)
    jbx_timings = JobexTimings.from_file(
        os.path.join(test_files, "jobex", "time.stat_riper")
    )
    assert len(jbx_timings.steps_timings) == 7
    assert jbx_timings.steps_timings[0]["step"] == "scf_energy"
    assert jbx_timings.steps_timings[0]["details"] is None
    assert jbx_timings.steps_timings[3]["step"] == "statpt"
    assert jbx_timings.steps_timings[3]["details"] == "statpt ended normally"
    assert jbx_timings.total_time() == pytest.approx(1.089)
    assert jbx_timings.total_time(time="real") == pytest.approx(1.089)
    assert jbx_timings.total_time(time="user") == pytest.approx(1.032)
    assert jbx_timings.total_time(time="sys") == pytest.approx(0.045)
    assert jbx_timings.total_time(time="sys", step="nonexistent") == pytest.approx(0.0)
    assert jbx_timings.total_time(time="real", step="scf_energy") == pytest.approx(
        1.044
    )
    assert jbx_timings.total_time(time="user", step="statpt") == pytest.approx(0.034)

    # time.stat file from "jobex -ri -level cc2 -time" (relaxation with MP2)
    jbx_timings = JobexTimings.from_file(
        os.path.join(test_files, "jobex", "time.stat_mp2")
    )
    assert jbx_timings.steps_timings[0]["user"] == pytest.approx(1030.948)
    assert jbx_timings.total_time() == pytest.approx(491.805)
    assert jbx_timings.total_time("user") == pytest.approx(2595.512)
    assert len(jbx_timings.steps_timings) == 7

    # time.stat file from standard "jobex -time" (relaxation with dscf)
    jbx_timings = JobexTimings.from_file(
        os.path.join(test_files, "jobex", "time.stat_dscf")
    )
    assert len(jbx_timings.steps_timings) == 7
    assert jbx_timings.total_time(step="scf_energy") == pytest.approx(0.531)

    # fake time.stat file with hours
    jbx_timings = JobexTimings.from_file(
        os.path.join(test_files, "jobex", "time.stat_hours")
    )
    assert len(jbx_timings.steps_timings) == 3
    assert jbx_timings.total_time(step="scf_energy") == pytest.approx(12926.528)
    assert jbx_timings.total_time(time="user", step="scf_energy") == pytest.approx(
        454320.507
    )
    assert jbx_timings.total_time(time="sys", step="statpt") == pytest.approx(
        136800.002
    )

    # fake time.stat with more than 2 lines for the step type and details
    with pytest.raises(
        RuntimeError, match=r"More than two time\.stat lines for a given jobex step"
    ):
        JobexTimings.from_file(os.path.join(test_files, "jobex", "time.stat_raise"))


def test_get_define_parameters():
    with pytest.raises(
        RuntimeError,
        match=r"Should provide at least one of "
        r'"define_template" or "define_parameters"',
    ):
        get_define_parameters(define_template=None, define_parameters=None)



# Part of atomate2-turbomole package.

"""Module containing utilities for the atomate2-turbomole package."""

from __future__ import annotations

import re
from datetime import datetime

from monty.json import MSONable
from turbomoleio.input.utils import get_define_template

float_number_re = r"[+-]?[0-9]*[.]?[0-9]+"
timing_re = r"(\d+h|)(\d+m)(" + float_number_re + r"s)"


class JobexTimings(MSONable):
    """Object containing timing information about a jobex run.

    Note: this will be moved and adapted in turbomoleio.
    """

    def __init__(self, steps_timings):
        """Construct JobexTimings.

        Args:
            steps_timings (list): List of the timings for each step.
        """
        self.steps_timings = steps_timings

    @classmethod
    def from_file(cls, filepath="time.stat"):
        """Create JobexTimings from file.

        Args:
            filepath: Path to the "time.stat" file containing timing information.

        Returns
        -------
            JobexTimings instance with the timing information of the jobex calculation.
        """
        with open(filepath) as f:
            string = f.read().strip()

        pattern = (
            r"([\s\S]*?\s+)"  # lazy matching
            r"real\s+"
            + timing_re
            + r"\s+user\s+"
            + timing_re
            + r"\s+sys\s+"
            + timing_re
        )
        match = re.findall(pattern, string)
        steps_timings = []

        for step in match:
            real_t = float(step[3][:-1]) + 60.0 * float(step[2][:-1])
            if step[1]:
                real_t += 3600.0 * float(step[1][:-1])
            user_t = float(step[6][:-1]) + 60.0 * float(step[5][:-1])
            if step[4]:
                user_t += 3600.0 * float(step[4][:-1])
            sys_t = float(step[9][:-1]) + 60.0 * float(step[8][:-1])
            if step[7]:
                sys_t += 3600.0 * float(step[7][:-1])

            sp = [line.strip() for line in step[0].strip().splitlines()]
            if len(sp) == 1:
                details = None
            elif len(sp) == 2:
                details = sp[1]
            else:
                raise RuntimeError(
                    "More than two time.stat lines for a given jobex step"
                )
            step_type = sp[0]
            steps_timings.append(
                {
                    "step": step_type,
                    "details": details,
                    "real": real_t,
                    "user": user_t,
                    "sys": sys_t,
                }
            )
        return cls(steps_timings=steps_timings)

    def total_time(self, time="real", step=None):
        """Get the total time.

        Args:
            time: Which type of time to use.
            step: Which type of step to use.

        Returns
        -------
            float: Time used in seconds.
        """
        if step is not None:
            return sum([s[time] for s in self.steps_timings if s["step"] == step])
        return sum([s[time] for s in self.steps_timings])


def get_define_parameters(define_template=None, define_parameters=None):
    """Get the define parameters based on a template and additional define parameters.

    Args:
        define_template: Name of template to use as basis for define parameters.
        define_parameters: Parameters for turbomoleio's DefineRunner.
    """
    if define_template is None and define_parameters is None:
        raise RuntimeError(
            'Should provide at least one of "define_template" or "define_parameters"'
        )
    dp = get_define_template(define_template) if define_template else {}
    if define_parameters:
        dp.update(define_parameters)
    return dp


def datetime_str() -> str:
    """
    Get a string representation of the current time.

    Returns
    -------
    str
        The current time.
    """
    return str(datetime.utcnow())



"""Definition of base turbomole job makers."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field

import turbomoleio.output.files
from atomate2.turbomole.custodian.jobs import TMJob
from atomate2.turbomole.schemas.task import TaskDocument
from atomate2.turbomole.sets.base import (
    BaseTurbomoleInputGenerator,
)
from custodian import Custodian
from jobflow import Maker, Response, job

logger = logging.getLogger(__name__)

__all__ = ["BaseTurbomoleMaker"]


@dataclass
class BaseTurbomoleMaker(Maker):
    """Base Turbomole job maker."""

    input_set_generator: BaseTurbomoleInputGenerator
    tm_exec: str
    name: str = "base turbomole job"
    command_options: list = field(default_factory=list)
    handlers: list = field(default_factory=list)
    validators: list = field(default_factory=list)
    output_cls_str: str = "ScfOutput"

    def __post_init__(self):
        """Set the output_cls based on the output_cls_str."""
        self.output_cls = getattr(turbomoleio.output.files, self.output_cls_str)

    @job
    def make(self, prev_output=None):
        """Create the dscf job."""
        self.input_set_generator.get_input_set(prev_output=prev_output)

        command_options = self.get_command_options()
        custodian_job = TMJob(executable=self.tm_exec, options=command_options)
        custodian = Custodian(
            handlers=self.handlers,
            validators=self.validators,
            jobs=[custodian_job],
        )
        custodian.run()

        # Parse the outputs
        if self.tm_exec == "jobex":
            output_file = "job.last"
        else:
            output_file = custodian_job.output_file
        doc = TaskDocument.from_directory(
            ".", output_file=output_file, output_cls=self.output_cls
        )

        return Response(output=doc)

    def get_command_options(self):
        """Get the options for the Turbomole executable."""
        return self.command_options



"""Definition of core Turbomole flow makers."""
from dataclasses import dataclass, field
from typing import Optional, Union

from atomate2.turbomole.jobs.core import (
    DefineMaker,
    DscfMaker,
    JobexMaker,
    RidftMaker,
    RiperMaker,
)
from jobflow import Flow, Maker
from pymatgen.core.structure import Molecule, Structure
from turbomoleio.core.molecule import MoleculeSystem
from turbomoleio.core.periodic import PeriodicSystem


def check_periodicity_consistence(
    scf_maker: Union[DscfMaker, RidftMaker, RiperMaker] = None,
    system: Union[Molecule, MoleculeSystem, Structure, PeriodicSystem] = None,
):
    """Check the system periodicity and the consistency of the maker.

    If system is Molecule or MoleculeSystem, self.scf_maker can only be Dscf or Ridft,
    else if system is a Structure or a PeriodicSystem, self.scf_maker must be Riper.

    Args:
        scf_maker: maker object from atomate2-turbomole
        system: pymatgen Molecule/Structure object or turbomoleio
            MoleculeSystem/PeriodicSystem object.

    Return:
        None
    """
    if isinstance(system, (Molecule, MoleculeSystem)):
        if not isinstance(scf_maker, (DscfMaker, RidftMaker)):
            raise RuntimeError(
                "Inconsistent choice for make() in your Flow:\n"
                "scf_maker for Molecule/MoleculeSystem\n"
                "must be DscfMaker or RidftMaker.\n"
                f"Here the scf_maker is:\n{scf_maker}"
            )
    elif isinstance(system, (Structure, PeriodicSystem)):  # pragma: no branch (trivial)
        if not isinstance(scf_maker, RiperMaker):
            raise RuntimeError(
                "Inconsistent choice for make() in your Flow:\n"
                "scf_maker for Structure/PeriodicSystem\n"
                "must be RiperMaker.\n"
                f"Here the scf_maker is:\n{scf_maker}"
            )
    else:  # pragma: no cover (trivial)
        raise RuntimeError(
            "System can only be Molecule, MoleculeSystem,"
            "Structure or PeriodicSystem.\n"
            f"Here the system is of type {type(system)}"
        )


@dataclass
class ScfFlowMaker(Maker):
    """
    Maker for a define + scf flow.

    Args:
        define_maker: Maker for define.
        scf_maker: Maker for dscf/ridft/riper job.
    """

    name: str = "Scf Flow Maker"
    define_maker: DefineMaker = field(
        default_factory=lambda: DefineMaker.from_define_template("ridft")
    )
    scf_maker: Union[DscfMaker, RidftMaker, RiperMaker] = field(
        default_factory=lambda: RidftMaker()
    )

    def make(
        self,
        system,
        charge=None,
        unpaired_electrons=None,
        prev_output=None,
    ):
        """
        Make a define+scf flow.

        Parameters
        ----------
        system
            either a pymatgen Molecule or Structure or a
            turbomoleio MoleculeSystem or PeriodicSystem.
        charge
            charge of the system.
        unpaired_electrons
            number of unpaired electrons.
        prev_output
            a previous output to start from.

        Returns
        -------
            Flow: jobflow Flow performing define and scf sequentially.
        """
        check_periodicity_consistence(self.scf_maker, system)

        if prev_output is None:
            define_job = self.define_maker.make(
                system, charge=charge, unpaired_electrons=unpaired_electrons
            )
            scf_job = self.scf_maker.make(prev_output=define_job.output.dir_name)
            jobs = [define_job, scf_job]
        else:
            scf_job = self.scf_maker.make(prev_output=prev_output)
            jobs = [scf_job]

        flow = Flow(jobs, output=scf_job.output)
        return flow

    @classmethod
    def dscf(cls, define_parameters=None, **kwargs):
        """Create Scf flow using dscf."""
        return cls(
            define_maker=DefineMaker.from_define_template(
                "dscf", define_parameters=define_parameters
            ),
            scf_maker=DscfMaker(),
            **kwargs,
        )

    @classmethod
    def ridft(cls, define_parameters=None, **kwargs):
        """Create Scf flow using ridft."""
        return cls(
            define_maker=DefineMaker.from_define_template(
                "ridft", define_parameters=define_parameters
            ),
            scf_maker=RidftMaker(),
            **kwargs,
        )

    @classmethod
    def riper(cls, define_parameters=None, **kwargs):
        """Create Scf flow using riper."""
        return cls(
            define_maker=DefineMaker.from_define_template(
                "ridft", define_parameters=define_parameters
            ),
            scf_maker=RiperMaker(),
            **kwargs,
        )


@dataclass
class JobexFlowMaker(Maker):
    """
    Maker for a define + jobex flow.

    Args:
        define_maker: Maker for define.
        jobex_maker: Maker for jobex job.
        scf_maker: Maker for scf job.

    Options:
        max_cycles: max number of Jobex (geometry relaxation) cycles. If not provided,
            the value from the jobex_maker is used (by default: 1000). If both
            jobex_maker and max_cycles are provided, max_cycles takes precedence.
        pre_scf:
            if True:
                The flow will be: ScfFlow + JobexJob
            else:
                The flow will be: DefineJob + JobexJob

        post_scf:
            if True:
                The flow will end with a ScfJob.
                Since Jobex = [dscf, grad, statpt], the final energy will not be
                exactly the same. With this option on, a final scf is performed
                so that the energy in control is rewritten consistently.
            else:
                The final energy in control is the one from the N-1 step, while
                the one in Jobex output (geometry step N) is slightly different.
    """

    name: str = "Jobex Flow Maker"
    define_maker: DefineMaker = field(
        default_factory=lambda: DefineMaker.from_define_template("ridft")
    )
    max_cycles: Optional[int] = None
    jobex_maker: JobexMaker = field(default_factory=lambda: JobexMaker(max_cycles=1000))
    scf_maker: Union[DscfMaker, RidftMaker, RiperMaker] = field(
        default_factory=lambda: RidftMaker()
    )
    pre_scf_flow_maker: ScfFlowMaker = field(default_factory=lambda: ScfFlowMaker())
    pre_scf: bool = True
    post_scf: bool = False

    def __post_init__(self):
        """Post init modification of inner makers."""
        if self.max_cycles is not None:
            self.jobex_maker.max_cycles = self.max_cycles

    def make(
        self,
        system,
        charge=None,
        unpaired_electrons=None,
        prev_output=None,
    ):
        """
        Make a define + jobex + final scf flow.

        Parameters
        ----------
        system
            either a pymatgen Molecule or Structure or a
            turbomoleio MoleculeSystem or PeriodicSystem.
        charge
            charge of the system.
        unpaired_electrons
            number of unpaired electrons.
        prev_output
            a previous output to start from.

        Returns
        -------
            Flow: jobflow Flow performing sequentially:
                - a define job
                - a jobex job (with or without --riper option for periodic systems)
                - a final scf job (dscf or ridft) to make the output energy consistent
        """
        check_periodicity_consistence(self.scf_maker, system)

        if prev_output is None:  # pragma: no branch (trivial)
            if self.pre_scf:  # pragma: no cover (to be done)
                init_scf_flow = self.pre_scf_flow_maker.make(
                    system, charge=charge, unpaired_electrons=unpaired_electrons
                )
                jobex_job = self.jobex_maker.make(
                    prev_output=init_scf_flow.output.dir_name
                )
                jobs = [init_scf_flow, jobex_job]
            else:
                define_job = self.define_maker.make(
                    system, charge=charge, unpaired_electrons=unpaired_electrons
                )
                jobex_job = self.jobex_maker.make(
                    prev_output=define_job.output.dir_name
                )
                jobs = [define_job, jobex_job]
        else:  # pragma: no cover (to be done)
            jobex_job = self.jobex_maker.make(prev_output=prev_output)
            jobs = [jobex_job]

        if self.post_scf:  # pragma: no cover (to be done)
            post_scf_job = self.scf_maker.make(prev_output=jobex_job)
            jobs.append(post_scf_job)
            flow_output = post_scf_job.output
        else:
            flow_output = jobex_job.output

        # TODO: see if we want to store more than just the final job
        #  i.e. only the post scf if it's there or also the jobex and
        #  possibly the pre-dscf/ridft and define jobs.
        flow = Flow(jobs, output=flow_output)
        return flow

    @classmethod
    def dscf(cls, define_parameters=None, **kwargs):
        """Create Scf flow using dscf."""
        pre_scf_flow_maker = ScfFlowMaker.dscf(define_parameters=define_parameters)
        return cls(
            define_maker=DefineMaker.from_define_template(
                "dscf", define_parameters=define_parameters
            ),
            pre_scf_flow_maker=pre_scf_flow_maker,
            scf_maker=DscfMaker(),
            **kwargs,
        )

    @classmethod
    def ridft(cls, define_parameters=None, **kwargs):
        """Create Scf flow using ridft."""
        pre_scf_flow_maker = ScfFlowMaker.ridft(define_parameters=define_parameters)
        return cls(
            define_maker=DefineMaker.from_define_template(
                "ridft", define_parameters=define_parameters
            ),
            pre_scf_flow_maker=pre_scf_flow_maker,
            scf_maker=RidftMaker(),
            **kwargs,
        )

    @classmethod
    def riper(cls, define_parameters=None, **kwargs):
        """Create Scf flow using riper."""
        pre_scf_flow_maker = ScfFlowMaker.riper(define_parameters=define_parameters)
        return cls(
            define_maker=DefineMaker.from_define_template(
                "ridft", define_parameters=define_parameters
            ),
            pre_scf_flow_maker=pre_scf_flow_maker,
            scf_maker=RiperMaker(),
            **kwargs,
        )



"""Module defining Turbomole calculation types."""

from emmet.core.utils import ValueEnum


class RunType(ValueEnum):
    """Turbomole calculation run types."""


class TaskType(ValueEnum):
    """Turbomole calculation task types."""


class CalcType(ValueEnum):
    """Turbomole calculation types."""



# # Part of turbomole package.
#
# """Tests of the database tools in the turbomole package."""
#
# import os
#
# import turbomole.database
# from turbomole.database import ResultsDatabase
#
#
# def test_database(mocker, test_files, mongo_collection):
#     mock = mocker.patch.object(turbomole.database, "MongoClient", autospec=True)
#     rdb = ResultsDatabase(
#         host="localhost",
#         port=1234,
#         database="results_db",
#         collection_name="projectA_results",
#         user="johndoe",
#         password="password",
#     )
#     mock.assert_called_once()
#     rdb.insert({"a": 1})
#     rdb.collection.insert_one.assert_called_once()
#     rdb.collection.insert_one.assert_called_with({"a": 1})
#
#     mock.reset_mock()
#     rdb = ResultsDatabase.from_db_file(
#         os.path.join(test_files, "misc", "results_db.yaml")
#     )
#     assert rdb.user == "testuser"
#     assert rdb.password == "testpassword"
#     assert rdb.host == "somehost"
#     assert rdb.port == 12345
#     assert rdb.database == "results"
#     rdb.db.__getitem__.assert_called_once()  # == 'projectA_results'
#     rdb.db.__getitem__.assert_called_with("projectA_results")
#
#     mocker.patch.object(turbomole.database, "ResultsDatabase", autospec=True)
#     rdb = ResultsDatabase(
#         host="localhost",
#         port=1234,
#         database="results_db",
#         collection_name="projectA_results",
#         user="johndoe",
#         password="password",
#     )
#     rdb.collection = mongo_collection
#     rdb.insert({"b": "b"})
#     assert rdb.collection.count_documents({}) == 1
#     rdb.insert({"c": "c"})
#     assert rdb.collection.count_documents({}) == 2



"""Module defining core atomate2-turbomole input set generators."""
from __future__ import annotations

import json
from dataclasses import dataclass, field

import numpy as np
from atomate2.turbomole.sets.base import (
    BaseTurbomoleInputGenerator,
    DefineInputSet,
    TurbomoleInputSet,
)
from atomate2.turbomole.utils import get_define_parameters
from pymatgen.core import Molecule, Structure
from turbomoleio.core.control import cpc
from turbomoleio.core.molecule import MoleculeSystem
from turbomoleio.core.periodic import PeriodicSystem


@dataclass
class TurbomoleDefineInputGenerator(BaseTurbomoleInputGenerator):
    """A class to generate Turbomole input sets."""

    define_template: str | None = None
    define_parameters: dict = field(default_factory=dict)
    user_define_parameters_settings: dict = field(default_factory=dict)
    user_datagroups_settings: dict = field(default_factory=dict)
    use_system_charge: bool = True
    use_system_multiplicity: bool = False

    def get_input_set(
        self,
        system: MoleculeSystem | PeriodicSystem | Molecule | Structure,
        charge=None,
        unpaired_electrons=None,
    ) -> DefineInputSet:
        """
        Generate a DefineInputSet object.

        Parameters
        ----------
        system
            either a pymatgen Molecule or Structure or a
            turbomoleio MoleculeSystem or PeriodicSystem.
        charge
            charge of the system.
        unpaired_electrons
            number of unpaired electrons.

        Returns
        -------
            a DefineInputSet object

        TODO: can we use pydantic to make sure system is of the correct kind?
        """
        # Define the system
        if isinstance(system, Molecule):
            system = MoleculeSystem(system)
        elif isinstance(system, Structure):
            system = PeriodicSystem(system)
        elif not isinstance(  # pragma: no branch (trivial)
            system, (MoleculeSystem, PeriodicSystem)
        ):
            raise RuntimeError(
                'The "molecule_or_structure" input should be a pymatgen '
                "Molecule/Structure or a turbomoleio MoleculeSystem/PeriodicSystem."
            )

        # Get the define parameters
        dp = get_define_parameters(self.define_template, self.define_parameters)
        if unpaired_electrons is not None:
            dp["unpaired_electrons"] = unpaired_electrons
        elif self.use_system_multiplicity:  # pragma: no branch (trivial)
            raise NotImplementedError(
                "multiplicity from the system not yet implemented."
            )
        if charge is not None:
            dp["charge"] = charge
        elif dp.get("charge", None) is None and self.use_system_charge:
            dp["charge"] = system._molecule_or_structure.charge
        # Charges should be integers
        if dp.get("charge", None) is not None:
            int_charge = int(np.around(dp["charge"]))
            if not np.isclose(int_charge, dp["charge"]):
                raise ValueError("Charge is not close to an integer.")
            dp["charge"] = int_charge

        # The exchange-correlation functional should be set outside of define as it
        # does not know about libxc functionals
        func = dp.pop("functional", None)

        return DefineInputSet(
            inputs={
                "coord": system.to_coord_string(),
                "define_parameters.json": json.dumps(dp),
                "datagroups.json": json.dumps(self.user_datagroups_settings),
            },
            xc_func=func,
            datagroups=self.user_datagroups_settings,
            define_parameters=dp,
            system=system,
        )


@dataclass
class TurbomoleInputGenerator(BaseTurbomoleInputGenerator):
    """A class to generate Turbomole input sets."""

    def get_input_set(self, prev_output=None) -> TurbomoleInputSet:
        """
        Generate a TurbomoleInputSet object.

        TODO: what kind of file is prev_output?
        """
        if prev_output is not None:
            cpc(".", control_dir=prev_output)
        return TurbomoleInputSet()


@dataclass
class DscfInputGenerator(TurbomoleInputGenerator):
    """A class to generate Turbomole input sets."""

    def get_input_set(self, prev_output=None) -> TurbomoleInputSet:
        """
        Generate a TurbomoleInputSet object.

        TODO: what kind of file is prev_output?
        """
        if prev_output is not None:
            cpc(".", control_dir=prev_output)
        return TurbomoleInputSet()



"""Module defining base atomate2-turbomole input set and generator."""
import logging
from dataclasses import dataclass

from pymatgen.io.core import InputGenerator, InputSet

__all__ = [
    "DefineInputSet",
    "TurbomoleInputSet",
    "BaseTurbomoleInputGenerator",
]

logger = logging.getLogger(__name__)


class DefineInputSet(InputSet):
    """A class to represent a set of Turbomole's define inputs."""

    pass


class TurbomoleInputSet(InputSet):
    """A class to represent a set of Turbomole inputs."""

    pass


@dataclass
class BaseTurbomoleInputGenerator(InputGenerator):
    """Base input generator for Turbolole."""

    pass



"""Core definitions of Turbomole calculation documents."""

from typing import Any, Dict, List, Union

from atomate2.turbomole.schemas.calc_types import CalcType, RunType, TaskType
from jobflow.utils import ValueEnum
from pydantic import BaseModel, Field
from pymatgen.core.structure import Molecule, Structure


class Status(ValueEnum):
    """Turbomole calculation state."""

    SUCCESS = "successful"
    FAILED = "failed"


class TurbomoleObject(ValueEnum):
    """Types of Turbomole data objects."""

    ENERGY = "energy"


class CalculationInput(BaseModel):
    """Summary of inputs for a Turbomole calculation."""

    structure: Union[Structure, Molecule] = Field(
        None, description="The input structure/molecule object"
    )

    turbomole_input: Dict = Field(
        None, description="The Turbomole input used for this task"
    )

    dft: Dict = Field(
        None,
        description="DFT parameters used in the last calc of this task.",
    )

    turbomole_global: Dict = Field(
        None,
        description="Turbomole global parameters used in the last calc of this task.",
    )


class CalculationOutput(BaseModel):
    """Document defining Turbomole calculation outputs."""

    energy: float = Field(
        None, description="The final total DFT energy for the calculation"
    )
    energy_per_atom: float = Field(
        None, description="The final DFT energy per atom for the calculation"
    )
    structure: Union[Structure, Molecule] = Field(
        None, description="The final structure/molecule from the calculation"
    )
    efermi: float = Field(
        None, description="The Fermi level from the calculation in eV"
    )
    is_metal: bool = Field(None, description="Whether the system is metallic")
    bandgap: float = Field(None, description="The band gap from the calculation in eV")
    cbm: float = Field(
        None,
        description="The conduction band minimum in eV (if system is not metallic)",
    )
    vbm: float = Field(
        None, description="The valence band maximum in eV (if system is not metallic)"
    )
    ionic_steps: List[Dict[str, Any]] = Field(
        None, description="Energy, forces, and structure for each ionic step"
    )


class Calculation(BaseModel):
    """Full Turbomole calculation inputs and outputs."""

    dir_name: str = Field(
        None, description="The directory for this Turbomole calculation"
    )
    turbomole_version: str = Field(
        None, description="Turbomole version used to perform the calculation"
    )
    has_completed: Status = Field(
        None, description="Whether Turbomole completed the calculation successfully"
    )
    input: CalculationInput = Field(
        None, description="Turbomole input settings for the calculation"
    )
    output: CalculationOutput = Field(
        None, description="The Turbomole calculation output"
    )
    completed_at: str = Field(
        None, description="Timestamp for when the calculation was completed"
    )
    task_name: str = Field(
        None, description="Name of task given by custodian (e.g., relax1, relax2)"
    )
    output_file_paths: Dict[str, str] = Field(
        None,
        description="Paths (relative to dir_name) of the Turbomole output files "
        "associated with this calculation",
    )
    # bader: Dict = Field(None, description="Output from the bader software")
    run_type: RunType = Field(
        None, description="Calculation run type (e.g., HF, HSE06, PBE)"
    )
    task_type: TaskType = Field(
        None, description="Calculation task type (e.g., Structure Optimization)."
    )
    calc_type: CalcType = Field(
        None, description="Return calculation type (run type + task_type)."
    )



# Part of atomate2-turbomole package.

"""Module containing the custodian jobs for the atomate2-turbomole package."""

import shlex
import subprocess
from pathlib import Path

from custodian.custodian import Job


class TMJob(Job):
    """A job to run one of the TM executables."""

    def __init__(self, executable, output_file=None, stderr_file=None, options=None):
        """Construct the custodian Job.

        Args:
            executable: Turbomole executable.
            output_file: Filename to use for the stdout.
            stderr_file: Filename to use for the stderr.
            options: Options for jobex (e.g. "-rijk", ["-level", "cc2"], "-time", ...).
        """
        self.executable = executable
        self.output_file = output_file or Path(self.executable).name + ".out"
        self.stderr_file = stderr_file or Path(self.executable).name + ".err"
        if options is None:
            self.options = []
        elif isinstance(options, str):
            self.options = shlex.split(options)
        elif isinstance(options, list):
            self.options = list(options) or []
        else:
            raise TypeError(
                f'"options" should be a string, a list or None, '
                f"while it is {type(options)}"
            )

    def setup(self, directory="./"):
        """Set up the custodian job. Nothing at this stage."""
        pass  # pragma: no cover

    def run(self, directory="./"):
        """
        Run the selected executable.

        Returns
        -------
            a Popen process
        """
        cmd = [self.executable, *self.options]

        with open(self.output_file, "w") as f_out, open(
            self.stderr_file, "w", buffering=1
        ) as f_err:
            p = subprocess.Popen(cmd, stdout=f_out, stderr=f_err)
        return p

    def postprocess(self, directory="./"):
        """Postprocess the custodian job. Nothing at this stage."""
        pass  # pragma: no cover

    @classmethod
    def dscf(cls, **kwargs):
        """Create custodian TMJob for dscf."""
        return cls(executable="dscf", **kwargs)

    @classmethod
    def ridft(cls, **kwargs):
        """Create custodian TMJob for ridft."""
        return cls(executable="ridft", **kwargs)

    @classmethod
    def jobex(cls, jobex_time=True, **kwargs):
        """Create custodian TMJob for jobex."""
        options = kwargs.pop("options", [])
        if isinstance(options, str):
            options = options = shlex.split(options)
        if jobex_time and "-time" not in options:
            options.append("-time")
        if not jobex_time and "-time" in options:
            raise ValueError(
                "Explicitly asked jobex_time=False while -time option is set."
            )
        return cls(executable="jobex", options=options, **kwargs)

    @classmethod
    def ricc2(cls, **kwargs):
        """Create custodian TMJob for ricc2."""
        return cls(executable="ricc2", **kwargs)

    @classmethod
    def riper(cls, **kwargs):
        """Create custodian TMJob for riper."""
        return cls(executable="riper", **kwargs)

    @classmethod
    def statpt(cls, **kwargs):
        """Create custodian TMJob for statpt."""
        return cls(executable="statpt", **kwargs)



"""Definition of core Turbomole job makers."""
import copy
import os
import shutil
from dataclasses import dataclass, field

from atomate2.turbomole.custodian.validators import (
    JobexGeoOptConvergedValidator,
    ScfConvergedValidator,
)
from atomate2.turbomole.jobs.base import BaseTurbomoleMaker
from atomate2.turbomole.schemas.task import DefineTaskDocument
from atomate2.turbomole.sets.core import (
    BaseTurbomoleInputGenerator,
    TurbomoleDefineInputGenerator,
    TurbomoleInputGenerator,
)
from jobflow import Maker, Response, job
from monty.os import cd, makedirs_p
from turbomoleio import DefineRunner, MoleculeSystem
from turbomoleio.core.control import cdg, mdgo, sdg


@dataclass
class DefineMaker(Maker):
    """Base maker for Turbomole's define jobs."""

    input_set_generator: TurbomoleDefineInputGenerator = field(
        default_factory=lambda: TurbomoleDefineInputGenerator(define_template="ridft")
    )
    name: str = "define"
    metric_options = (3, 2, 1, -1, -2, -3)
    define_timeout = 5

    @classmethod
    def from_define_template(cls, define_template, define_parameters=None):
        """Initialize the define maker using a define template."""
        # TODO: allow tuning the rest also ?
        define_parameters = define_parameters or {}
        return cls(
            TurbomoleDefineInputGenerator(
                define_template=define_template, define_parameters=define_parameters
            )
        )

    @job
    def make(self, system, charge=None, unpaired_electrons=None):
        """Create define job for a given system.

        Parameters
        ----------
        system
            either a pymatgen Molecule or Structure or a
            turbomoleio MoleculeSystem or PeriodicSystem.
        charge
            charge of the system.
        unpaired_electrons
            number of unpaired electrons.

        Returns
        -------
            a jobflow Job to run define with the desired options for the given system,
            charge and number of unpaired electrons.
        """
        dis = self.input_set_generator.get_input_set(
            system, charge=charge, unpaired_electrons=unpaired_electrons
        )
        if not isinstance(dis.system, MoleculeSystem):
            raise NotImplementedError()

        # Run define
        try:
            self.run_define(define_input_set=dis)
        except RuntimeError:  # pragma: no cover (tricky to test)
            return Response(stop_jobflow=True)

        # Update datagroups and functional
        self.update_datagroups_functional(
            datagroups=dis.datagroups, functional=dis.xc_func
        )

        # Parse the outputs
        doc = DefineTaskDocument.from_directory(".")

        return Response(output=doc)

    def run_define(self, define_input_set):
        """Run define."""
        define_parameters = define_input_set.define_parameters
        dr_succeeded = False
        define_parameters = copy.deepcopy(define_parameters)

        for metric in self.metric_options:  # pragma: no branch (trivial)
            metric_dir = f"define_metric_{metric}"
            makedirs_p(metric_dir)
            try:
                with cd(metric_dir):
                    define_input_set.write_input(directory=".")
                    define_parameters["metric"] = metric
                    dr = DefineRunner(parameters=define_parameters, timeout=60)
                    dr.run_full()
                dr_succeeded = True
            except BaseException:  # pragma: no cover (tricky)
                pass

            if dr_succeeded:  # pragma: no branch (trivial)
                # Move the files of the succeeded define run to the main directory
                files = os.listdir(metric_dir)
                for f in files:
                    if os.path.exists(f):  # pragma: no cover (tricky)
                        raise FileExistsError(
                            f'Will not copy "{f}" file as it already exists.'
                        )
                    shutil.move(os.path.join(metric_dir, f), f)
                break

        if not dr_succeeded:  # pragma: no cover (tricky)
            raise RuntimeError(
                "Running define went wrong with all the different metrics."
            )

        return metric

    def update_datagroups_functional(self, datagroups, functional):
        """Update the datagroups and functional."""
        if datagroups:  # pragma: no cover (trivial)
            for dg, db in datagroups.items():
                cdg(dg, db)

        # Setup exchange-correlation functional
        if functional:  # pragma: no branch (trivial)
            if functional == "b97m-v":  # pragma: no cover (trivial)
                mdgo("dft", {"functional": "functional libxc 254"})
            else:
                mdgo("dft", {"functional": f"functional {functional}"})
            # DO the Self-Consistent Non-Local (doscnl) density-based
            # vanderwaals corrections. This is needed to compute the gradients.
            if functional in [  # pragma: no cover (trivial)
                "wb97x-v",
                "wb97m-v",
                "b97m-v",
            ]:
                cdg("doscnl", "")


@dataclass
class DscfMaker(BaseTurbomoleMaker):
    """Base maker for dscf jobs."""

    input_set_generator: BaseTurbomoleInputGenerator = field(
        default_factory=TurbomoleInputGenerator
    )
    tm_exec: str = "dscf"
    name: str = "dscf"
    handlers: list = field(default_factory=list)
    validators: list = field(
        default_factory=lambda: [ScfConvergedValidator(output_file="dscf.out")]
    )


@dataclass
class RidftMaker(BaseTurbomoleMaker):
    """Base maker for ridft jobs."""

    input_set_generator: BaseTurbomoleInputGenerator = field(
        default_factory=TurbomoleInputGenerator
    )
    tm_exec: str = "ridft"
    name: str = "ridft"
    handlers: list = field(default_factory=list)
    validators: list = field(
        default_factory=lambda: [ScfConvergedValidator(output_file="ridft.out")]
    )


@dataclass
class RiperMaker(BaseTurbomoleMaker):
    """Base maker for riper jobs."""

    input_set_generator: BaseTurbomoleInputGenerator = field(
        default_factory=TurbomoleInputGenerator
    )
    tm_exec: str = "riper"
    name: str = "riper"
    handlers: list = field(default_factory=list)
    validators: list = field(
        default_factory=lambda: [ScfConvergedValidator(output_file="riper.out")]
    )


@dataclass
class JobexMaker(BaseTurbomoleMaker):
    """Base maker for jobex jobs."""

    input_set_generator: BaseTurbomoleInputGenerator = field(
        default_factory=TurbomoleInputGenerator
    )
    tm_exec: str = "jobex"
    name: str = "jobex"
    command_options: list = field(default_factory=list)
    handlers: list = field(default_factory=list)
    validators: list = field(default_factory=lambda: [JobexGeoOptConvergedValidator()])
    max_cycles: int = 100
    output_cls_str: str = "JobexOutput"

    def get_command_options(self):  # pragma: no cover (to be done)
        """Get the options for the jobex executable."""
        command_options = list(self.command_options)
        rij = sdg("rij")
        if rij is not None:
            command_options.append("-ri")
        rik = sdg("rik")
        if rik is not None:
            command_options.append("-rijk")
        if self.max_cycles is not None:
            command_options.extend(["-c", f"{self.max_cycles}"])
        periodic = sdg("periodic")
        if periodic is not None:
            if int(periodic.strip()) in (1, 2, 3):
                command_options.append("-riper")
            else:
                raise ValueError(
                    "periodic should not be present in the control file "
                    "or be one of 1, 2 or 3."
                )
        return command_options



"""Core definition of atomate2-turbomole task document."""
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from atomate2.turbomole.schemas.calculation import (
    Calculation,
    Status,
)
from atomate2.turbomole.utils import datetime_str
from emmet.core.math import Matrix3D, Vector3D
from emmet.core.structure import MoleculeMetadata
from monty.os import cd
from pydantic import BaseModel, Field
from pymatgen.core.structure import Molecule, Structure
from turbomoleio.core.control import Control, Energy, Gradient
from turbomoleio.core.molecule import MoleculeSystem
from turbomoleio.core.periodic import PeriodicSystem
from turbomoleio.output.data import RunData
from turbomoleio.output.files import BaseData, JobexOutput, ScfOutput
from turbomoleio.output.states import States

__all__ = [
    "DefineTaskDocument",
    "TaskDocument",
]

logger = logging.getLogger(__name__)
_DefineTD_T = TypeVar("_DefineTD_T", bound="DefineTaskDocument")
_TD_T = TypeVar("_TD_T", bound="TaskDocument")


# TODO: This was a gigantic copy paste from atomate2
# Cp2kInput is defined in pymatgen, TM Inputs are not defined in TurbomoleIO


class DefineTaskDocument(MoleculeMetadata):
    """Definition of define task document."""

    dir_name: str = Field(None, description="The directory for this define task")
    last_updated: str = Field(
        default_factory=datetime_str,
        description="Timestamp for this task document was last updated",
    )
    completed_at: str = Field(
        None, description="Timestamp for when this task was completed"
    )
    output: dict = Field(None, description="The output of the final calculation")

    # molecule: Molecule = Field(
    #     None, description="Final output molecule from the task"
    # )

    @classmethod
    def from_directory(
        cls: Type[_DefineTD_T],
        dir_name: Union[Path, str],
    ) -> _DefineTD_T:
        """
        Create a task document from a directory containing the output of a define run.

        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.

        Returns
        -------
        DefineTaskDocument
            A task document for the calculation.
        """
        logger.info(f"Getting task doc in: {dir_name}")

        dir_name = Path(dir_name).absolute()

        with cd(dir_name):
            system = MoleculeSystem.from_file("control", fmt="coord")
            molecule = system.molecule
            control = Control.from_file("control")
            rundata = RunData.from_file("define.log")

        # Remove dummy atoms
        molecule.remove_species(["x", "X", "q", "Q"])

        doc = cls.from_molecule(meta_molecule=molecule)
        ddict = doc.dict()
        data = {
            "molecule": molecule,
            "system": system,
            "dir_name": str(dir_name),
            "completed_at": str(rundata.end_time),
            "output": {"control": control},
        }
        ddict.update(data)
        doc = cls(**ddict)
        # doc = doc.copy(update=data)
        return doc


class InputSummary(BaseModel):
    """Summary of inputs for a Turbomole calculation."""

    structure: Union[Structure, Molecule] = Field(
        None, description="The input structure object"
    )

    xc: str = Field(
        None, description="Exchange-correlation functional used if not the default"
    )

    @classmethod
    def from_turbomole_calc_doc(  # pragma: no cover (not used)
        cls, calc_doc: Calculation
    ) -> "InputSummary":
        """
        Create calculation input summary from a calculation document.

        Parameters
        ----------
        calc_doc
            A Turbomole calculation document.

        Returns
        -------
        InputSummary
            A summary of the input structure and parameters.
        """
        summary = None

        return cls(
            structure=calc_doc.input.structure,
            xc=str(calc_doc.run_type),
            summary=summary,
        )


class OutputSummary(BaseModel):
    """Summary of the outputs for a Turbomole calculation."""

    structure: Optional[Union[Structure, Molecule]] = Field(
        None, description="The output structure object"
    )
    energy: Optional[float] = Field(
        None, description="The final total DFT energy for the last calculation"
    )
    energy_per_atom: Optional[float] = Field(
        None, description="The final DFT energy per atom for the last calculation"
    )
    bandgap: Optional[float] = Field(
        None, description="The DFT bandgap for the last calculation"
    )
    cbm: Optional[float] = Field(None, description="CBM for this calculation")
    vbm: Optional[float] = Field(None, description="VBM for this calculation")
    forces: Optional[List[Vector3D]] = Field(
        None, description="Forces on atoms from the last calculation"
    )
    stress: Optional[Matrix3D] = Field(
        None, description="Stress on the unit cell from the last calculation"
    )

    @classmethod
    def from_turbomole_calc_doc(  # pragma: no cover (not used)
        cls, calc_doc: Calculation
    ) -> "OutputSummary":
        """
        Create a summary of Turbomole calculation outputs.

        Parameters
        ----------
        calc_doc
            A Turbomole calculation document.

        Returns
        -------
        OutputSummary
            The calculation output summary.
        """
        if calc_doc.output.ionic_steps:
            forces = calc_doc.output.ionic_steps[-1].get("forces", None)
            stress = calc_doc.output.ionic_steps[-1].get("stress", None)
        else:
            forces = None
            stress = None
        return cls(
            structure=calc_doc.output.structure,
            energy=calc_doc.output.energy,
            energy_per_atom=calc_doc.output.energy_per_atom,
            bandgap=calc_doc.output.bandgap,
            cbm=calc_doc.output.cbm,
            vbm=calc_doc.output.vbm,
            forces=forces,
            stress=stress,
        )


class TaskDocument(MoleculeMetadata):
    """Definition of Turbomole task document."""

    dir_name: str = Field(None, description="The directory for this Turbomole task")
    last_updated: str = Field(
        default_factory=datetime_str,
        description="Timestamp for this task document was last updated",
    )
    completed_at: str = Field(
        None, description="Timestamp for when this task was completed"
    )
    input: InputSummary = Field(None, description="The input to the first calculation")
    output: OutputSummary = Field(
        None, description="The output of the final calculation"
    )
    structure: Union[Structure, Molecule] = Field(
        None, description="Final output structure from the task"
    )
    state: Status = Field(None, description="State of this task")
    task_label: str = Field(None, description="A description of the task")
    tags: Optional[List[str]] = Field(
        None, description="Metadata tags for this task document"
    )
    author: Optional[str] = Field(
        None, description="Author extracted from transformations"
    )

    """
    TODO: enforce on real cases
    """
    identifier: Optional[dict] = Field(
        None,
        description="Identifier dictionary for known databases "
        "such as Materials Project {'type': 'MP', 'value': 'mp-7500'}",
    )

    calcs_reversed: List[Calculation] = Field(
        default_factory=list,
        description="The inputs and outputs for all Turbomole runs in this task.",
    )
    custodian: Any = Field(
        None,
        description="Information on the custodian settings used to run this "
        "calculation, parsed from a custodian.json file",
    )

    @classmethod
    def from_directory(
        cls: Type[_TD_T],
        dir_name: Union[Path, str],
        output_file: str,
        additional_fields: Dict[str, Any] = None,
        output_cls: BaseData = ScfOutput,
    ) -> _TD_T:
        """
        Create a task document from a directory containing the output of a define run.

        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.
        output_file
            The output file
        additional_fields
            Dictionary of additional fields to add to output document.
        output_cls
            Class used to parse the output file.

        Returns
        -------
        ScfTaskDocument
            A task document for the calculation.

        Args:
            output_cls:
        """
        logger.info(f"Getting task doc in: {dir_name}")

        additional_fields = {} if additional_fields is None else additional_fields
        dir_name = Path(dir_name).absolute()

        with cd(dir_name):
            control = Control.from_file("control")
            output = output_cls.from_file(output_file)
            energy_output = output.energy if isinstance(output, JobexOutput) else output
            scf_states = States.from_file("control")
            energies = Energy.from_file()
            try:
                gradient = Gradient.from_file()
            except FileNotFoundError:
                gradient = None

            periodic = control.show_data_group("periodic")

            if periodic is None:  # pragma: no branch
                attr = "from_molecule"
                ms = MoleculeSystem.from_file("control", fmt="coord")
                molecule = ms.molecule.copy()
                # Remove dummy atoms
                molecule.remove_species(["x", "X", "q", "Q"])
                dat = {
                    "system": ms,
                    "structure": molecule,
                    "meta_molecule": molecule,
                    "input": InputSummary(
                        structure=molecule, xc=energy_output.dft.functional.name
                    ),
                }

            elif int(periodic.strip()) in (
                1,
                2,
                3,
            ):  # pragma: no cover (not implemented)
                attr = "from_structure"
                ps = PeriodicSystem.from_file("control", fmt="coord")
                dat = {
                    "system": ps,
                    "structure": ps.structure,
                    "meta_structure": ps.structure,
                }

            else:  # pragma: no cover (trivial)
                raise ValueError(
                    "periodic should not be present in the control file "
                    "or be one of 1, 2 or 3."
                )
            dat["completed_at"] = str(energy_output.run.end_time)
            dat["state"] = "successful"
            dat["task_label"] = output_cls.__name__

        doc = getattr(cls, attr)(**dat)
        ddict = doc.dict()
        data = {
            "dir_name": str(dir_name),
            "output": {
                "control": control,
                "output": output,
                "energies": energies,
                "gradient": gradient,
                "states": scf_states,
            },
            "energy": energies.total[-1],
            "homo": scf_states.homo,
            "lumo": scf_states.lumo,
        }
        ddict.update(data)
        doc = cls(**ddict)
        # doc = doc.copy(update=data)
        doc = doc.copy(update=additional_fields)
        return doc



# Part of atomate2-turbomole package.

"""General tests of the atomate2-turbomole package."""


from atomate2.turbomole import __version__


def test_version():
    assert isinstance(__version__, str)
